#include "Gedeng/Logger.h"
#include "Gedeng/ParticleSystem.h"
#include "Gedeng/QuadMesh.h"
#include "Gedeng/TextLabel.h"
#include "Gedeng/Vector3.h"
#include <GLFW/glfw3.h>
#define GEDENG_MAIN
#include <Gedeng.h>

class ParticleApp : public Gedeng::Application {
  public:
    ParticleApp(unsigned long ms_per_update, unsigned int window_size_x, unsigned int window_size_y,
                Gedeng::String window_name)
        : Application(ms_per_update, window_size_x, window_size_y, window_name), number_of_steps(10.0),
          number_of_refinement_steps(10.0), bump_depth(0.1),
          render_shader(Gedeng::Shader("Shader/bump.vs", "Shader/bump.fs")),
          camera(Gedeng::FPSCamera(90, 1920, 1080, 0.1, 1000.0)),
          albedo("Resources/Textures/PavingStones/PavingStones070_2K_Color.jpg", Gedeng::Texture::Settings()),
          bump("Resources/Textures/PavingStones/PavingStones070_2K_Displacement.jpg", Gedeng::Texture::Settings()),
          normal("Resources/Textures/PavingStones/PavingStones070_2K_Normal.jpg", Gedeng::Texture::Settings()),
          quad_mesh(Gedeng::QuadMesh(10.0)) {
        particles.set_position(glm::vec3(0.0f, 0.0f, -10.0f));
        particles.set_velocity(glm::vec3(-2, 4, -2), glm::vec3(2, 6, 2));
        particles.set_gravity(glm::vec3(0, -4, 0));
        particles.set_color(glm::vec3(0.0f, 0.5f, 1.0f));
        particles.set_lifetime(1.0f, 1.5f);
        particles.set_size(0.1);
        particles.set_interval(0.05f);
        particles.set_number_of_particles(1);
        particles.set_texture(&albedo);

        camera.translate(glm::vec3(0.0, 2.0, 1.0));
        // camera.rotate(30, glm::vec3(1.0, 0.0, 0.0));
    }

    ~ParticleApp() = default;

    void fixed_update(double delta) override {
        // camera.rotate(delta * 180, glm::vec3(0.0, 1.0, 0.0));
        camera.update(delta);

        if (Gedeng::Input::is_mouse_down(GLFW_MOUSE_BUTTON_LEFT))
            particles.set_position(camera.get_view_ray().get_plane_collision().collision_position);
    }

    void dynamic_update(double delta) override {
        glClearColor(0.1, 0.1f, 0.1, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        render_shader.use();

        // Camera
        render_shader.setMat4("projection", camera.get_projection());
        render_shader.setMat4("view", camera.get_view());
        render_shader.setVec3("viewPos", camera.get_translation());

        // Lighting
        render_shader.setVec3("lightPos", glm::vec3(0.0, 1.0, 5.0));

        render_shader.setFloat("number_of_steps", glm::max(0.0f, number_of_steps));
        render_shader.setFloat("number_of_refinement_steps", glm::max(0.0f, number_of_refinement_steps));
        render_shader.setFloat("bump_depth", glm::max(0.0f, bump_depth));

        // Textures
        albedo.bind_to(0);
        normal.bind_to(1);
        bump.bind_to(2);

        // Quad which is rendered onto
        quad_mesh.render(render_shader);

        // Particles
        particles.set_camera(camera);
        particles.update(delta);
        particles.render();
    }

  private:
    float number_of_steps;
    float number_of_refinement_steps;
    float bump_depth;

    Gedeng::Shader render_shader;

    Gedeng::VertexBuffer vertex_rectangle;

    Gedeng::FPSCamera camera;

    Gedeng::Texture albedo;
    Gedeng::Texture bump;
    Gedeng::Texture normal;

    Gedeng::QuadMesh quad_mesh;

    Gedeng::ParticleSystem particles;
};

Gedeng::Application *Gedeng::create_application() {
    GG_CLIENT_INFO("Creating Application");
    return new ParticleApp(20, 1920, 1080, String("Parallax Demo"));
}
